'use strict';

var fs = require('fs');
var JSZip = require('jszip');
var jexl = require('jexl');
var FileSaver = require('file-saver');

/*! *****************************************************************************
Copyright (c) Microsoft Corporation.

Permission to use, copy, modify, and/or distribute this software for any
purpose with or without fee is hereby granted.

THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
PERFORMANCE OF THIS SOFTWARE.
***************************************************************************** */

function __awaiter(thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
}

/**
 * isNodeJS
 * @see https://github.com/mozilla/pdf.js/blob/master/src/shared/is_node.js
 */
const isNodeJS = typeof process === 'object' &&
    process + '' === '[object process]' &&
    !process.versions.nw &&
    // @ts-expect-error TS2339: Property 'type' does not exist on type 'Process'.
    !(process.versions.electron && process.type && process.type !== 'browser');
/**
 * global/window
 */
const g = new Function('return this')();
/**
 * Avoid complaints
 */
//
// if (typeof g.require === 'undefined') {
//     g.require = function(obj) { return obj; };
// }

const origin = isNodeJS ? require('@xmldom/xmldom') : g;
origin.DOMImplementation;
const DOMParser = origin.DOMParser;
const XMLSerializer = origin.XMLSerializer;

const constants = {
    delimiters: {
        start: '{{',
        end: '}}'
    },
    empty: '',
    locale: 'en-US',
    mimeType: 'application/xml',
};

/**
 * ar-EG
 */
const culture$c = {
    name: "ar-EG",
    numberFormat: {
        pattern: ["-n"],
        decimals: 2,
        ",": ",",
        ".": ".",
        groupSize: [3],
        percent: {
            pattern: ["-n%", "n%"],
            decimals: 2,
            ",": ",",
            ".": ".",
            groupSize: [3],
            symbol: "%"
        },
        currency: {
            name: "Egyptian Pound",
            abbr: "EGP",
            pattern: ["-n $", "n $"],
            decimals: 2,
            ",": ",",
            ".": ".",
            groupSize: [3],
            symbol: "ج.م.‏"
        }
    },
    calendars: {
        standard: {
            days: {
                names: ["الأحد", "الإثنين", "الثلاثاء", "الأربعاء", "الخميس", "الجمعة", "السبت"],
                namesAbbr: ["الأحد", "الإثنين", "الثلاثاء", "الأربعاء", "الخميس", "الجمعة", "السبت"],
                namesShort: ["ح", "ن", "ث", "ر", "خ", "ج", "س"]
            },
            months: {
                names: ["يناير", "فبراير", "مارس", "أبريل", "مايو", "يونيو", "يوليو", "أغسطس", "سبتمبر", "أكتوبر", "نوفمبر", "ديسمبر"],
                namesAbbr: ["يناير", "فبراير", "مارس", "أبريل", "مايو", "يونيو", "يوليو", "أغسطس", "سبتمبر", "أكتوبر", "نوفمبر", "ديسمبر"]
            },
            AM: ["ص", "ص", "ص"],
            PM: ["م", "م", "م"],
            patterns: {
                d: "dd/MM/yyyy",
                D: "dd MMMM, yyyy",
                F: "dd MMMM, yyyy hh:mm:ss tt",
                g: "dd/MM/yyyy hh:mm tt",
                G: "dd/MM/yyyy hh:mm:ss tt",
                m: "dd MMMM",
                M: "dd MMMM",
                s: "yyyy'-'MM'-'dd'T'HH':'mm':'ss",
                t: "hh:mm tt",
                T: "hh:mm:ss tt",
                u: "yyyy'-'MM'-'dd HH':'mm':'ss'Z'",
                y: "MMMM, yyyy",
                Y: "MMMM, yyyy"
            },
            "/": "/",
            ":": ":",
            firstDay: 6
        }
    }
};

/**
 * bn-IN
 */
const culture$b = {
    name: "bn-IN",
    numberFormat: {
        pattern: ["-n"],
        decimals: 2,
        ",": ",",
        ".": ".",
        groupSize: [3, 2],
        percent: {
            pattern: ["-n%", "n%"],
            decimals: 2,
            ",": ",",
            ".": ".",
            groupSize: [3, 2],
            symbol: "%"
        },
        currency: {
            name: "Indian Rupee",
            abbr: "INR",
            pattern: ["$ -n", "$ n"],
            decimals: 2,
            ",": ",",
            ".": ".",
            groupSize: [3, 2],
            symbol: "₹"
        }
    },
    calendars: {
        standard: {
            days: {
                names: ["রবিবার", "সোমবার", "মঙ্গলবার", "বুধবার", "বৃহস্পতিবার", "শুক্রবার", "শনিবার"],
                namesAbbr: ["রবি.", "সোম.", "মঙ্গল.", "বুধ.", "বৃহস্পতি.", "শুক্র.", "শনি."],
                namesShort: ["র", "স", "ম", "বু", "বৃ", "শু", "শ"]
            },
            months: {
                names: ["জানুয়ারী", "ফেব্রুয়ারী", "মার্চ", "এপ্রিল", "মে", "জুন", "জুলাই", "আগস্ট", "সেপ্টেম্বর", "অক্টোবর", "নভেম্বর", "ডিসেম্বর"],
                namesAbbr: ["জানু.", "ফেব্রু.", "মার্চ", "এপ্রিল", "মে", "জুন", "জুলাই", "আগ.", "সেপ্টে.", "অক্টো.", "নভে.", "ডিসে."]
            },
            AM: ["AM", "am", "AM"],
            PM: ["PM", "pm", "PM"],
            patterns: {
                d: "dd-MM-yy",
                D: "dd MMMM yyyy",
                F: "dd MMMM yyyy HH.mm.ss",
                g: "dd-MM-yy HH.mm",
                G: "dd-MM-yy HH.mm.ss",
                m: "d MMMM",
                M: "d MMMM",
                s: "yyyy'-'MM'-'dd'T'HH':'mm':'ss",
                t: "HH.mm",
                T: "HH.mm.ss",
                u: "yyyy'-'MM'-'dd HH':'mm':'ss'Z'",
                y: "MMMM, yyyy",
                Y: "MMMM, yyyy"
            },
            "/": "-",
            ":": ".",
            firstDay: 1
        }
    }
};

/**
 * de-DE
 */
const culture$a = {
    name: "de-DE",
    numberFormat: {
        pattern: ["-n"],
        decimals: 2,
        ",": ".",
        ".": ",",
        groupSize: [3],
        percent: {
            pattern: ["-n %", "n %"],
            decimals: 2,
            ",": ".",
            ".": ",",
            groupSize: [3],
            symbol: "%"
        },
        currency: {
            name: "Euro",
            abbr: "EUR",
            pattern: ["-n $", "n $"],
            decimals: 2,
            ",": ".",
            ".": ",",
            groupSize: [3],
            symbol: "€"
        }
    },
    calendars: {
        standard: {
            days: {
                names: ["Sonntag", "Montag", "Dienstag", "Mittwoch", "Donnerstag", "Freitag", "Samstag"],
                namesAbbr: ["So", "Mo", "Di", "Mi", "Do", "Fr", "Sa"],
                namesShort: ["So", "Mo", "Di", "Mi", "Do", "Fr", "Sa"]
            },
            months: {
                names: ["Januar", "Februar", "März", "April", "Mai", "Juni", "Juli", "August", "September", "Oktober", "November", "Dezember"],
                namesAbbr: ["Jan", "Feb", "Mrz", "Apr", "Mai", "Jun", "Jul", "Aug", "Sep", "Okt", "Nov", "Dez"]
            },
            AM: [""],
            PM: [""],
            patterns: {
                d: "dd.MM.yyyy",
                D: "dddd, d. MMMM yyyy",
                F: "dddd, d. MMMM yyyy HH:mm:ss",
                g: "dd.MM.yyyy HH:mm",
                G: "dd.MM.yyyy HH:mm:ss",
                m: "d. MMMM",
                M: "d. MMMM",
                s: "yyyy'-'MM'-'dd'T'HH':'mm':'ss",
                t: "HH:mm",
                T: "HH:mm:ss",
                u: "yyyy'-'MM'-'dd HH':'mm':'ss'Z'",
                y: "MMMM yyyy",
                Y: "MMMM yyyy"
            },
            "/": ".",
            ":": ":",
            firstDay: 1
        }
    }
};

/**
 * en-GB
 */
const culture$9 = {
    name: "en-GB",
    numberFormat: {
        pattern: ["-n"],
        decimals: 2,
        ",": ",",
        ".": ".",
        groupSize: [3],
        percent: {
            pattern: ["-n%", "n%"],
            decimals: 2,
            ",": ",",
            ".": ".",
            groupSize: [3],
            symbol: "%"
        },
        currency: {
            name: "British Pound",
            abbr: "GBP",
            pattern: ["-$n", "$n"],
            decimals: 2,
            ",": ",",
            ".": ".",
            groupSize: [3],
            symbol: "£"
        }
    },
    calendars: {
        standard: {
            days: {
                names: ["Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"],
                namesAbbr: ["Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"],
                namesShort: ["Su", "Mo", "Tu", "We", "Th", "Fr", "Sa"]
            },
            months: {
                names: ["January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"],
                namesAbbr: ["Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"]
            },
            AM: ["AM", "am", "AM"],
            PM: ["PM", "pm", "PM"],
            patterns: {
                d: "dd/MM/yyyy",
                D: "dd MMMM yyyy",
                F: "dd MMMM yyyy HH:mm:ss",
                g: "dd/MM/yyyy HH:mm",
                G: "dd/MM/yyyy HH:mm:ss",
                m: "d MMMM",
                M: "d MMMM",
                s: "yyyy'-'MM'-'dd'T'HH':'mm':'ss",
                t: "HH:mm",
                T: "HH:mm:ss",
                u: "yyyy'-'MM'-'dd HH':'mm':'ss'Z'",
                y: "MMMM yyyy",
                Y: "MMMM yyyy"
            },
            "/": "/",
            ":": ":",
            firstDay: 1
        }
    }
};

/**
 * en-US
 */
const culture$8 = {
    name: 'en-US',
    numberFormat: {
        pattern: ['-n'],
        decimals: 2,
        ',': ',',
        '.': '.',
        groupSize: [3],
        percent: {
            pattern: ['-n %', 'n %'],
            decimals: 2,
            ',': ',',
            '.': '.',
            groupSize: [3],
            symbol: '%'
        },
        currency: {
            name: 'US Dollar',
            abbr: 'USD',
            pattern: ['($n)', '$n'],
            decimals: 2,
            ',': ',',
            '.': '.',
            groupSize: [3],
            symbol: '$'
        }
    },
    calendars: {
        standard: {
            days: {
                names: ['Sunday', 'Monday', 'Tuesday', 'Wednesday', 'Thursday', 'Friday', 'Saturday'],
                namesAbbr: ['Sun', 'Mon', 'Tue', 'Wed', 'Thu', 'Fri', 'Sat'],
                namesShort: ['Su', 'Mo', 'Tu', 'We', 'Th', 'Fr', 'Sa']
            },
            months: {
                names: ['January', 'February', 'March', 'April', 'May', 'June', 'July', 'August', 'September', 'October', 'November', 'December'],
                namesAbbr: ['Jan', 'Feb', 'Mar', 'Apr', 'May', 'Jun', 'Jul', 'Aug', 'Sep', 'Oct', 'Nov', 'Dec']
            },
            AM: ['AM', 'am', 'AM'],
            PM: ['PM', 'pm', 'PM'],
            patterns: {
                d: 'M/d/yyyy',
                D: 'dddd, MMMM dd, yyyy',
                F: 'dddd, MMMM dd, yyyy h:mm:ss tt',
                g: 'M/d/yyyy h:mm tt',
                G: 'M/d/yyyy h:mm:ss tt',
                m: 'MMMM dd',
                M: 'MMMM dd',
                s: "yyyy'-'MM'-'ddTHH':'mm':'ss",
                t: 'h:mm tt',
                T: 'h:mm:ss tt',
                u: "yyyy'-'MM'-'dd HH':'mm':'ss'Z'",
                y: 'MMMM, yyyy',
                Y: 'MMMM, yyyy'
            },
            '/': '/',
            ':': ':',
            firstDay: 0,
            twoDigitYearMax: 2029
        }
    }
};

/**
 * es-ES
 */
const culture$7 = {
    name: "es-ES",
    numberFormat: {
        pattern: ["-n"],
        decimals: 2,
        ",": ".",
        ".": ",",
        groupSize: [3],
        percent: {
            pattern: ["-n %", "n %"],
            decimals: 2,
            ",": ".",
            ".": ",",
            groupSize: [3],
            symbol: "%"
        },
        currency: {
            name: "Euro",
            abbr: "EUR",
            pattern: ["-n $", "n $"],
            decimals: 2,
            ",": ".",
            ".": ",",
            groupSize: [3],
            symbol: "€"
        }
    },
    calendars: {
        standard: {
            days: {
                names: ["domingo", "lunes", "martes", "miércoles", "jueves", "viernes", "sábado"],
                namesAbbr: ["do.", "lu.", "ma.", "mi.", "ju.", "vi.", "sá."],
                namesShort: ["D", "L", "M", "X", "J", "V", "S"]
            },
            months: {
                names: ["enero", "febrero", "marzo", "abril", "mayo", "junio", "julio", "agosto", "septiembre", "octubre", "noviembre", "diciembre"],
                namesAbbr: ["ene.", "feb.", "mar.", "abr.", "may.", "jun.", "jul.", "ago.", "sep.", "oct.", "nov.", "dic."]
            },
            AM: [""],
            PM: [""],
            patterns: {
                d: "dd/MM/yyyy",
                D: "dddd, d' de 'MMMM' de 'yyyy",
                F: "dddd, d' de 'MMMM' de 'yyyy H:mm:ss",
                g: "dd/MM/yyyy H:mm",
                G: "dd/MM/yyyy H:mm:ss",
                m: "d 'de' MMMM",
                M: "d 'de' MMMM",
                s: "yyyy'-'MM'-'dd'T'HH':'mm':'ss",
                t: "H:mm",
                T: "H:mm:ss",
                u: "yyyy'-'MM'-'dd HH':'mm':'ss'Z'",
                y: "MMMM' de 'yyyy",
                Y: "MMMM' de 'yyyy"
            },
            "/": "/",
            ":": ":",
            firstDay: 1
        }
    }
};

/**
 * fr-FR
 */
const culture$6 = {
    name: "fr-FR",
    numberFormat: {
        pattern: ["-n"],
        decimals: 2,
        ",": " ",
        ".": ",",
        groupSize: [3],
        percent: {
            pattern: ["-n %", "n %"],
            decimals: 2,
            ",": " ",
            ".": ",",
            groupSize: [3],
            symbol: "%"
        },
        currency: {
            name: "Euro",
            abbr: "EUR",
            pattern: ["-n $", "n $"],
            decimals: 2,
            ",": " ",
            ".": ",",
            groupSize: [3],
            symbol: "€"
        }
    },
    calendars: {
        standard: {
            days: {
                names: ["dimanche", "lundi", "mardi", "mercredi", "jeudi", "vendredi", "samedi"],
                namesAbbr: ["dim.", "lun.", "mar.", "mer.", "jeu.", "ven.", "sam."],
                namesShort: ["di", "lu", "ma", "me", "je", "ve", "sa"]
            },
            months: {
                names: ["janvier", "février", "mars", "avril", "mai", "juin", "juillet", "août", "septembre", "octobre", "novembre", "décembre"],
                namesAbbr: ["janv.", "févr.", "mars", "avr.", "mai", "juin", "juil.", "août", "sept.", "oct.", "nov.", "déc."]
            },
            AM: [""],
            PM: [""],
            patterns: {
                d: "dd/MM/yyyy",
                D: "dddd d MMMM yyyy",
                F: "dddd d MMMM yyyy HH:mm:ss",
                g: "dd/MM/yyyy HH:mm",
                G: "dd/MM/yyyy HH:mm:ss",
                m: "d MMMM",
                M: "d MMMM",
                s: "yyyy'-'MM'-'dd'T'HH':'mm':'ss",
                t: "HH:mm",
                T: "HH:mm:ss",
                u: "yyyy'-'MM'-'dd HH':'mm':'ss'Z'",
                y: "MMMM yyyy",
                Y: "MMMM yyyy"
            },
            "/": "/",
            ":": ":",
            firstDay: 1
        }
    }
};

/**
 * hi-IN
 */
const culture$5 = {
    name: "hi-IN",
    numberFormat: {
        pattern: ["-n"],
        decimals: 2,
        ",": ",",
        ".": ".",
        groupSize: [3, 2],
        percent: {
            pattern: ["-n%", "n%"],
            decimals: 2,
            ",": ",",
            ".": ".",
            groupSize: [3, 2],
            symbol: "%"
        },
        currency: {
            name: "Indian Rupee",
            abbr: "INR",
            pattern: ["$ -n", "$n"],
            decimals: 2,
            ",": ",",
            ".": ".",
            groupSize: [3, 2],
            symbol: "₹"
        }
    },
    calendars: {
        standard: {
            days: {
                names: ["रविवार", "सोमवार", "मंगलवार", "बुधवार", "गुरुवार", "शुक्रवार", "शनिवार"],
                namesAbbr: ["रवि.", "सोम.", "मंगल.", "बुध.", "गुरु.", "शुक्र.", "शनि."],
                namesShort: ["र", "स", "म", "ब", "ग", "श", "श"]
            },
            months: {
                names: ["जनवरी", "फरवरी", "मार्च", "अप्रैल", "मई", "जून", "जुलाई", "अगस्त", "सितम्बर", "अक्तूबर", "नवम्बर", "दिसम्बर"],
                namesAbbr: ["जनवरी", "फरवरी", "मार्च", "अप्रैल", "मई", "जून", "जुलाई", "अगस्त", "सितम्बर", "अक्तूबर", "नवम्बर", "दिसम्बर"]
            },
            AM: ["पूर्वाह्न", "पूर्वाह्न", "पूर्वाह्न"],
            PM: ["अपराह्न", "अपराह्न", "अपराह्न"],
            patterns: {
                d: "dd-MM-yyyy",
                D: "dd MMMM yyyy",
                F: "dd MMMM yyyy HH:mm:ss",
                g: "dd-MM-yyyy HH:mm",
                G: "dd-MM-yyyy HH:mm:ss",
                m: "dd MMMM",
                M: "dd MMMM",
                s: "yyyy'-'MM'-'dd'T'HH':'mm':'ss",
                t: "HH:mm",
                T: "HH:mm:ss",
                u: "yyyy'-'MM'-'dd HH':'mm':'ss'Z'",
                y: "MMMM, yyyy",
                Y: "MMMM, yyyy"
            },
            "/": "-",
            ":": ":",
            firstDay: 1
        }
    }
};

/**
 * it-IT
 */
const culture$4 = {
    name: "it-IT",
    numberFormat: {
        pattern: ["-n"],
        decimals: 2,
        ",": ".",
        ".": ",",
        groupSize: [3],
        percent: {
            pattern: ["-n%", "n%"],
            decimals: 2,
            ",": ".",
            ".": ",",
            groupSize: [3],
            symbol: "%"
        },
        currency: {
            name: "Euro",
            abbr: "EUR",
            pattern: ["-n $", "n $"],
            decimals: 2,
            ",": ".",
            ".": ",",
            groupSize: [3],
            symbol: "€"
        }
    },
    calendars: {
        standard: {
            days: {
                names: ["domenica", "lunedì", "martedì", "mercoledì", "giovedì", "venerdì", "sabato"],
                namesAbbr: ["dom", "lun", "mar", "mer", "gio", "ven", "sab"],
                namesShort: ["do", "lu", "ma", "me", "gi", "ve", "sa"]
            },
            months: {
                names: ["gennaio", "febbraio", "marzo", "aprile", "maggio", "giugno", "luglio", "agosto", "settembre", "ottobre", "novembre", "dicembre"],
                namesAbbr: ["gen", "feb", "mar", "apr", "mag", "giu", "lug", "ago", "set", "ott", "nov", "dic"]
            },
            AM: [""],
            PM: [""],
            patterns: {
                d: "dd/MM/yyyy",
                D: "dddd d MMMM yyyy",
                F: "dddd d MMMM yyyy HH:mm:ss",
                g: "dd/MM/yyyy HH:mm",
                G: "dd/MM/yyyy HH:mm:ss",
                m: "d MMMM",
                M: "d MMMM",
                s: "yyyy'-'MM'-'dd'T'HH':'mm':'ss",
                t: "HH:mm",
                T: "HH:mm:ss",
                u: "yyyy'-'MM'-'dd HH':'mm':'ss'Z'",
                y: "MMMM yyyy",
                Y: "MMMM yyyy"
            },
            "/": "/",
            ":": ":",
            firstDay: 1
        }
    }
};

/**
 * ja-JP
 */
const culture$3 = {
    name: "ja-JP",
    numberFormat: {
        pattern: ["-n"],
        decimals: 2,
        ",": ",",
        ".": ".",
        groupSize: [3],
        percent: {
            pattern: ["-n%", "n%"],
            decimals: 2,
            ",": ",",
            ".": ".",
            groupSize: [3],
            symbol: "%"
        },
        currency: {
            name: "Japanese Yen",
            abbr: "JPY",
            pattern: ["-$n", "$n"],
            decimals: 0,
            ",": ",",
            ".": ".",
            groupSize: [3],
            symbol: "¥"
        }
    },
    calendars: {
        standard: {
            days: {
                names: ["日曜日", "月曜日", "火曜日", "水曜日", "木曜日", "金曜日", "土曜日"],
                namesAbbr: ["日", "月", "火", "水", "木", "金", "土"],
                namesShort: ["日", "月", "火", "水", "木", "金", "土"]
            },
            months: {
                names: ["1月", "2月", "3月", "4月", "5月", "6月", "7月", "8月", "9月", "10月", "11月", "12月"],
                namesAbbr: ["1", "2", "3", "4", "5", "6", "7", "8", "9", "10", "11", "12"]
            },
            AM: ["午前", "午前", "午前"],
            PM: ["午後", "午後", "午後"],
            patterns: {
                d: "yyyy/MM/dd",
                D: "yyyy'年'M'月'd'日'",
                F: "yyyy'年'M'月'd'日' H:mm:ss",
                g: "yyyy/MM/dd H:mm",
                G: "yyyy/MM/dd H:mm:ss",
                m: "M月d日",
                M: "M月d日",
                s: "yyyy'-'MM'-'dd'T'HH':'mm':'ss",
                t: "H:mm",
                T: "H:mm:ss",
                u: "yyyy'-'MM'-'dd HH':'mm':'ss'Z'",
                y: "yyyy'年'M'月'",
                Y: "yyyy'年'M'月'"
            },
            "/": "/",
            ":": ":",
            firstDay: 0
        }
    }
};

/**
 * pt-PT
 */
const culture$2 = {
    name: "pt-PT",
    numberFormat: {
        pattern: ["-n"],
        decimals: 2,
        ",": " ",
        ".": ",",
        groupSize: [3],
        percent: {
            pattern: ["-n%", "n%"],
            decimals: 2,
            ",": " ",
            ".": ",",
            groupSize: [3],
            symbol: "%"
        },
        currency: {
            name: "Euro",
            abbr: "EUR",
            pattern: ["-n $", "n $"],
            decimals: 2,
            ",": " ",
            ".": ",",
            groupSize: [3],
            symbol: "€"
        }
    },
    calendars: {
        standard: {
            days: {
                names: ["domingo", "segunda-feira", "terça-feira", "quarta-feira", "quinta-feira", "sexta-feira", "sábado"],
                namesAbbr: ["dom", "seg", "ter", "qua", "qui", "sex", "sáb"],
                namesShort: ["D", "S", "T", "Q", "Q", "S", "S"]
            },
            months: {
                names: ["janeiro", "fevereiro", "março", "abril", "maio", "junho", "julho", "agosto", "setembro", "outubro", "novembro", "dezembro"],
                namesAbbr: ["jan", "fev", "mar", "abr", "mai", "jun", "jul", "ago", "set", "out", "nov", "dez"]
            },
            AM: [""],
            PM: [""],
            patterns: {
                d: "dd/MM/yyyy",
                D: "d' de 'MMMM' de 'yyyy",
                F: "d' de 'MMMM' de 'yyyy HH:mm:ss",
                g: "dd/MM/yyyy HH:mm",
                G: "dd/MM/yyyy HH:mm:ss",
                m: "d 'de' MMMM",
                M: "d 'de' MMMM",
                s: "yyyy'-'MM'-'dd'T'HH':'mm':'ss",
                t: "HH:mm",
                T: "HH:mm:ss",
                u: "yyyy'-'MM'-'dd HH':'mm':'ss'Z'",
                y: "MMMM' de 'yyyy",
                Y: "MMMM' de 'yyyy"
            },
            "/": "/",
            ":": ":",
            firstDay: 0
        }
    }
};

/**
 * ru-RU
 */
const culture$1 = {
    name: "ru-RU",
    numberFormat: {
        pattern: ["-n"],
        decimals: 2,
        ",": " ",
        ".": ",",
        groupSize: [3],
        percent: {
            pattern: ["-n%", "n%"],
            decimals: 2,
            ",": " ",
            ".": ",",
            groupSize: [3],
            symbol: "%"
        },
        currency: {
            name: "Russian Ruble",
            abbr: "RUB",
            pattern: ["-n $", "n $"],
            decimals: 2,
            ",": " ",
            ".": ",",
            groupSize: [3],
            symbol: "₽"
        }
    },
    calendars: {
        standard: {
            days: {
                names: ["воскресенье", "понедельник", "вторник", "среда", "четверг", "пятница", "суббота"],
                namesAbbr: ["Вс", "Пн", "Вт", "Ср", "Чт", "Пт", "Сб"],
                namesShort: ["Вс", "Пн", "Вт", "Ср", "Чт", "Пт", "Сб"]
            },
            months: {
                names: ["Январь", "Февраль", "Март", "Апрель", "Май", "Июнь", "Июль", "Август", "Сентябрь", "Октябрь", "Ноябрь", "Декабрь"],
                namesAbbr: ["янв", "фев", "мар", "апр", "май", "июн", "июл", "авг", "сен", "окт", "ноя", "дек"]
            },
            AM: [""],
            PM: [""],
            patterns: {
                d: "dd.MM.yyyy",
                D: "d MMMM yyyy 'г.'",
                F: "d MMMM yyyy 'г.' H:mm:ss",
                g: "dd.MM.yyyy H:mm",
                G: "dd.MM.yyyy H:mm:ss",
                m: "d MMMM",
                M: "d MMMM",
                s: "yyyy'-'MM'-'dd'T'HH':'mm':'ss",
                t: "H:mm",
                T: "H:mm:ss",
                u: "yyyy'-'MM'-'dd HH':'mm':'ss'Z'",
                y: "MMMM yyyy",
                Y: "MMMM yyyy"
            },
            "/": ".",
            ":": ":",
            firstDay: 1
        }
    }
};

/**
 * zh-CN
 */
const culture = {
    name: "zh-CN",
    numberFormat: {
        pattern: ["-n"],
        decimals: 2,
        ",": ",",
        ".": ".",
        groupSize: [3],
        percent: {
            pattern: ["-n%", "n%"],
            decimals: 2,
            ",": ",",
            ".": ".",
            groupSize: [3],
            symbol: "%"
        },
        currency: {
            name: "Chinese Yuan",
            abbr: "CNY",
            pattern: ["$-n", "$n"],
            decimals: 2,
            ",": ",",
            ".": ".",
            groupSize: [3],
            symbol: "¥"
        }
    },
    calendars: {
        standard: {
            days: {
                names: ["星期日", "星期一", "星期二", "星期三", "星期四", "星期五", "星期六"],
                namesAbbr: ["周日", "周一", "周二", "周三", "周四", "周五", "周六"],
                namesShort: ["日", "一", "二", "三", "四", "五", "六"]
            },
            months: {
                names: ["一月", "二月", "三月", "四月", "五月", "六月", "七月", "八月", "九月", "十月", "十一月", "十二月"],
                namesAbbr: ["1月", "2月", "3月", "4月", "5月", "6月", "7月", "8月", "9月", "10月", "11月", "12月"]
            },
            AM: ["上午", "上午", "上午"],
            PM: ["下午", "下午", "下午"],
            patterns: {
                d: "yyyy/M/d",
                D: "yyyy'年'M'月'd'日'",
                F: "yyyy'年'M'月'd'日' H:mm:ss",
                g: "yyyy/M/d H:mm",
                G: "yyyy/M/d H:mm:ss",
                m: "M月d日",
                M: "M月d日",
                s: "yyyy'-'MM'-'dd'T'HH':'mm':'ss",
                t: "H:mm",
                T: "H:mm:ss",
                u: "yyyy'-'MM'-'dd HH':'mm':'ss'Z'",
                y: "yyyy'年'M'月'",
                Y: "yyyy'年'M'月'"
            },
            "/": "/",
            ":": ":",
            firstDay: 1
        }
    }
};

/**
 * cultureMap
 * Note: avoids circular references with YumTemplate
 */
const cultureMap = new Map([
    ['ar-EG', culture$c],
    ['bn-IN', culture$b],
    ['de-DE', culture$a],
    ['en-GB', culture$9],
    ['en-US', culture$8],
    ['es-ES', culture$7],
    ['fr-FR', culture$6],
    ['hi-IN', culture$5],
    ['it-IT', culture$4],
    ['ja-JP', culture$3],
    ['pt-PT', culture$2],
    ['ru-RU', culture$1],
    ['zh-CN', culture],
]);

/**
 * Error codes
 */
const errorCodes = new Map([
    // General errors
    [1000, { code: 1000, message: 'Unknown error' }],
    [1001, { code: 1001, message: 'Assertion error' }],
    // File errors
    [1010, { code: 1010, message: 'File not found' }],
    [1011, { code: 1011, message: 'Corrupted file' }],
    [1012, { code: 1012, message: 'File load error' }],
    // OOXML File format errors
    [1020, { code: 1020, message: 'Missing or malformed [Content_types].xml' }],
    // TODO Parsing Errors
    // JEXL Expression error (unless the expression engine is replaced)
    [1060, { code: 1060, message: 'Expression evaluation error' }],
    // if tag errors
    // 1100+
    // #each tag errors
    [1110, { code: 1110, message: 'The #each tag expects an array or undefined' }],
]);

/**
 * YumError
 */
class YumError extends Error {
    // TODO stack
    /**
     * constructor
     * @param code
     * @param options
     */
    constructor(code, options = {}) {
        const error = errorCodes.get(code) || { code, message: '' };
        if (!error.message) {
            error.message = `Missing error code ${code}`;
        }
        super(error.message); // 'Error' breaks prototype chain here
        this.name = 'YumError';
        // @see https://stackoverflow.com/questions/41102060/typescript-extending-error-class
        Object.setPrototypeOf(this, new.target.prototype); // restore prototype chain
        this.code = code;
        if (options.error instanceof Error) {
            this.originalError = options.error;
        }
        if (Object.prototype.toString.call(options.data) === '[object Object]') {
            this.data = options.data;
        }
        this.name = 'YumError';
    }
}

/**
 * Part
 */
class AbstractPart {
    /**
     * Constructor
     * @param name
     * @param type
     * @param xml
     * @param parent
     * @param options
     * @protected
     */
    constructor(name, type, xml, parent, options) {
        this.priority = 5;
        this._done = false;
        this._name = name;
        this._type = type;
        this._parent = parent; // The parent list of parts in YumTemplate
        this._options = options; // These options
        const ppXml = this._preProcess(xml);
        this._dom = new DOMParser().parseFromString(ppXml, constants.mimeType);
    }
    /**
     * name
     */
    get name() {
        return this._name;
    }
    /**
     * type
     */
    get type() {
        return this._type;
    }
    /**
     * done
     */
    get done() {
        return this._done;
    }
    /**
     * _preProcess
     * @param xml
     */
    _preProcess(xml) {
        // Do nothing by default
        return xml;
    }
    /**
     * serialize
     */
    serialize() {
        return new XMLSerializer().serializeToString(this._dom);
    }
}

/**
 * escapeRegExp
 * polyfill for RegExp.escape
 * @see https://github.com/tc39/proposal-regex-escaping/blob/main/polyfill.js
 * @param s
 */
function escapeRegExp(s) {
    return s.replace(/[\\^$*+?.()|[\]{}]/g, '\\$&');
}

/**
 * sanitizeWordMarkup
 *
 * Versioning in Word is tracked with rsidR sections, like:
 * {</w:t></w:r><w:r w:rsidR="00245996"><w:t>{</w:t></w:r><w:r><w:t>te</w:t></w:r><w:r w:rsidR="00130070"><w:t>x</w:t></w:r><w:r w:rsidR="00EE726C"><w:t>t</w:t></w:r><w:r w:rsidR="00060B35"><w:t>}</w:t></w:r><w:r><w:t>}
 * {{te</w:t></w:r><w:r w:rsidR="00130070"><w:t>x</w:t></w:r><w:r w:rsidR="00EE726C"><w:t>t</w:t></w:r><w:r w:rsidR="00F00E75"><w:t>}</w:t></w:r><w:r><w:t>}
 * {{</w:t></w:r><w:r w:rsidR="00BD72CA"><w:t>integer</w:t></w:r><w:r><w:t>}}
 * {{</w:t></w:r><w:r w:rsidR="00BD72CA"><w:t>b</w:t></w:r><w:r><w:t>oolean}}
 *
 * Proof errors are also marked:
 * {{<w:proofErr w:type="spellStart"/>boolean<w:proofErr w:type="spellEnd"/>}}
 *
 * @param xml
 * @param delimiters
 */
function sanitizeWordMarkup(xml, delimiters) {
    const { start, end } = delimiters;
    const garbage = '<\\/w:t><\\/w:r><w:r\\s?[^>]*><w:t\\s?[^>]*>';
    let str = '';
    for (let i = 0; i < start.length - 1; i++) {
        str += `${escapeRegExp(start.charAt(i))}(${garbage})?`;
    }
    str += `${start.slice(-1)}[^${escapeRegExp(end.charAt(0))}]+${end.slice(0, 1)}`;
    for (let i = 1; i < start.length; i++) {
        str += `(${garbage})?${escapeRegExp(end.charAt(i))}`;
    }
    // const rx_find = /{(<\/w:t><\/w:r><w:r\s?[^>]*><w:t\s?[^>]*>)?{[^{}]+}(<\/w:t><\/w:r><w:r\s?[^>]*><w:t\s?[^>]*>)?}/;
    const rx_find = new RegExp(str);
    const rx_replace = /<\/?w:[^>]+>/g;
    let pos1 = 0, pos2 = 0, ret = '';
    while (xml.length > 0) {
        // Search for an expression in xml
        pos1 = xml.search(rx_find);
        if (pos1 > -1) {
            // find the end of the expression
            pos2 = xml.indexOf('}', pos1 + 1);
            pos2 = xml.indexOf('}', pos2 + 1);
            ret += xml.substring(0, pos1);
            ret += xml.substring(pos1, pos2 + '}'.length).replace(rx_replace, '');
            xml = xml.substring(pos2 + '}'.length);
        }
        else {
            ret += xml;
            xml = '';
        }
    }
    return ret;
}

/**
 * MatchNode
 */
class MatchedNode {
    /**
     * constructor
     * @param node
     * @param match
     */
    constructor(node, match) {
        this.node = node;
        this._match = match;
    }
    /**
     * statement
     */
    get statement() {
        // @ts-expect-error TS2339: Property 'groups' does not exist on type 'string[]'.
        return this._match.groups.statement;
    }
    /**
     * expression
     */
    get expression() {
        // @ts-expect-error TS2339: Property 'groups' does not exist on type 'string[]'.
        return this._match.groups.expression;
    }
    /**
     * nodeValue
     */
    get nodeValue() {
        return this.node.nodeValue;
    }
    /**
     * replaceTag
     * @param data
     */
    replaceMatch(data) {
        const tag = this._match[0];
        const offset = (this.node.nodeValue || constants.empty).indexOf(tag);
        const count = tag.length;
        return this.node.replaceData(offset, count, data);
    }
}

/**
 * AbstractTag
 */
class AbstractTag {
    /**
     * constructor
     * @param matchedNode
     * @param parent
     * @protected
     */
    constructor(matchedNode, parent) {
        this.matchedNodes = new Map();
        this._done = false;
        this.children = [];
        this.addMatchedNode(matchedNode);
        this.parent = parent;
    }
    /**
     * addNode
     * @param matchedNode
     */
    addMatchedNode(matchedNode) {
        // TODO Also check that node.statement is part of Tag.statements
        this.matchedNodes.set(matchedNode.statement || constants.empty /* ExpressionTag.statement */, matchedNode);
    }
    /**
     * done
     */
    get done() {
        return this._done;
    }
}

/**
 * assert
 * @param condition
 * @param message
 */
function assert(condition, message) {
    if (!condition) {
        throw new YumError(1001, message ? { data: { message } } : undefined);
    }
}

/**
 * getCommonAncestor
 * @param nodes
 */
function getChildrenOfCommonAncestor(...nodes) {
    // @see https://stackoverflow.com/questions/3960843/how-to-find-the-nearest-common-ancestors-of-two-or-more-nodes
    const ret = [...nodes];
    if (!ret.every(node => node === ret[0])) {
        // If all nodes are not identical
        // Check each parentNode until they are all identical or root is reached
        while (ret.slice(1).every(node => node.parentNode !== ret[0].parentNode) &&
            ret.every(node => node.parentNode !== ret[0].ownerDocument)) {
            for (let i = 0; i < ret.length; i++) {
                ret[i] = ret[i].parentNode;
            }
        }
    }
    return ret;
}
/**
 * getSiblingsBetween
 * Note: beware! returns an empty array if node1 === node2 (or node2 prior to node1)
 * @param node1
 * @param node2
 */
function getSiblingsBetween(node1, node2) {
    assert(node1.parentNode === node2.parentNode);
    const ret = [];
    if (node1 !== node2) {
        let current = node1;
        // add siblings to ret array until we reach either node2 or null (last sibling)
        do {
            if (current.nextSibling) {
                current = current.nextSibling;
                ret.push(current);
            }
            else {
                break;
            }
        } while (current.nextSibling !== node2);
        // if we have reached the last sibling without reaching node2
        if (current.nextSibling !== node2) {
            ret.splice(0, ret.length); // empty
        }
    }
    return ret;
}
/**
 * contains
 * @param parent
 * @param child
 */
function contains(parent, child) {
    if (child === parent)
        return true;
    let current = child;
    while (current.parentNode) {
        if (current.parentNode === parent) {
            return true;
        }
        current = current.parentNode;
    }
    return false;
}

/**
 * ExpressionEngine
 */
class ExpressionEngine {
    /**
     * constructor
     * @param evaluate
     */
    constructor(evaluate) {
        this._evaluate = evaluate;
    }
    /**
     * setEval
     * @param evaluate
     */
    setEval(evaluate) {
        this._evaluate = evaluate;
    }
    /**
     * evaluate
     * @param expression
     * @param context
     */
    evaluate(expression, context) {
        return __awaiter(this, void 0, void 0, function* () {
            return yield this._evaluate(expression, context);
        });
    }
}
/**
 * Our expression engine
 */
// Functions
// Note: interestingly, there is no collision between function names and json property names,
// so the $ prefix could be discarded
jexl.addFunction('$min', Math.min);
jexl.addFunction('$max', Math.max);
// String transforms
jexl.addTransform('lower', (val) => String(val).toLowerCase());
jexl.addTransform('replace', (val, search, replacement) => String(val).replace(search, replacement));
jexl.addTransform('substr', (val, start, end) => String(val).substring(start, end));
jexl.addTransform('upper', (val) => String(val).toUpperCase());
// Boolean
// Number
// Date
// Array
const expressionEngine = new ExpressionEngine(jexl.eval.bind(jexl));

class EachTag extends AbstractTag {
    /**
     * constructor
     * @param matchedNode
     * @param parent
     */
    constructor(matchedNode, parent) {
        super(matchedNode, parent);
    }
    /**
     * render
     * @param data
     */
    render(data = {}) {
        return __awaiter(this, void 0, void 0, function* () {
            const eachMatchedNode = this.matchedNodes.get(EachTag.statement);
            const endMatchedNode = this.matchedNodes.get(EachTag.blocks[0]);
            if (eachMatchedNode && endMatchedNode) {
                // if #each and #endeach are part of the same node
                if (eachMatchedNode.node === endMatchedNode.node) ;
                else {
                    // Find the common parent node and child elements directly underneath,
                    // respectively containing #each and #endeach
                    const topNodes = getChildrenOfCommonAncestor(eachMatchedNode.node, endMatchedNode.node);
                    assert(topNodes[0].parentNode === topNodes[1].parentNode);
                    // Find siblings to repeat between #each and #endeach
                    const siblings = getSiblingsBetween(topNodes[0], topNodes[1]);
                    // Evaluate #each expression, which should be an array
                    let arr; // avoid Array<unknown> because a value which is not an array becomes new Array(value);
                    try {
                        arr = yield expressionEngine.evaluate(eachMatchedNode.expression, data);
                    }
                    catch (error) {
                        throw new YumError(1060, { error, data: { expression: eachMatchedNode.expression, data } });
                    }
                    if (typeof arr === 'undefined') {
                        arr = [];
                    }
                    else if (!Array.isArray(arr)) {
                        throw new YumError(1110, { data: { expression: eachMatchedNode.expression, data } });
                    }
                    const array = arr;
                    // Update xmldom and ast
                    const parent = topNodes[0].parentNode;
                    if (array.length === 0) {
                        for (const sibling of siblings) {
                            // Remove tags from ast
                            for (let i = this.children.length - 1; i >= 0; i--) {
                                const [matchedNode] = this.children[i].matchedNodes.values();
                                if (contains(sibling, matchedNode.node)) {
                                    this.children.splice(i, 1);
                                }
                            }
                            // remove nodes from DOM
                            parent.removeChild(sibling);
                        }
                    }
                    else {
                        for (let i = 1; i < array.length; i++) {
                            // Duplicate siblings for each array item
                            for (const sibling of siblings) {
                                const node = sibling.cloneNode(true);
                                parent.insertBefore(node, topNodes[1]);
                                // Parse new node
                                const _ast = new AbstractTag.TagParser(node).parse(); // TODO options with delimiters
                                // Add to document ast
                                this.children.push(..._ast);
                            }
                        }
                    }
                    // Remove #each and #endeach
                    topNodes.forEach(node => { parent.removeChild(node); });
                    // Process children
                    for (let i = 0; i < this.children.length; i++) {
                        yield this.children[i].render(array[Math.trunc(i / siblings.length)]);
                    }
                }
                this._done = true;
            }
        });
    }
}
EachTag.statement = '#each';
EachTag.blocks = ['#endeach'];

/**
 * ExpressionTag
 */
class ExpressionTag extends AbstractTag {
    /**
     * constructor
     * @param matchedNode
     * @param parent
     */
    constructor(matchedNode, parent) {
        super(matchedNode, parent);
    }
    /**
     * render
     * @param data
     */
    render(data) {
        return __awaiter(this, void 0, void 0, function* () {
            const matchedNode = this.matchedNodes.get(ExpressionTag.statement);
            assert(matchedNode instanceof MatchedNode);
            let str;
            try {
                // TODO what if condition is not a string???
                str = (yield expressionEngine.evaluate(matchedNode.expression, data));
            }
            catch (error) {
                throw new YumError(1060, { error, data: { expression: matchedNode.expression, data } });
            }
            matchedNode.replaceMatch(str);
            this._done = true;
        });
    }
}
ExpressionTag.statement = constants.empty;
ExpressionTag.blocks = [];

class IfTag extends AbstractTag {
    /**
     * constructor
     * @param matchedNode
     * @param parent
     */
    constructor(matchedNode, parent) {
        super(matchedNode, parent);
    }
    /**
     * render
     * @param data
     */
    render(data = {}) {
        return __awaiter(this, void 0, void 0, function* () {
            const ifMatchedNode = this.matchedNodes.get(IfTag.statement);
            const elseMatchedNode = this.matchedNodes.get(IfTag.blocks[0]);
            const endMatchedNode = this.matchedNodes.get(IfTag.blocks[1]);
            if (ifMatchedNode && endMatchedNode) {
                // Find the common parent node and child elements directly underneath,
                // respectively containing #if, #else and #endif
                const topNodes = elseMatchedNode instanceof MatchedNode ?
                    getChildrenOfCommonAncestor(ifMatchedNode.node, elseMatchedNode.node, endMatchedNode.node) :
                    getChildrenOfCommonAncestor(ifMatchedNode.node, endMatchedNode.node);
                assert(topNodes[0] === topNodes[topNodes.length - 1] || topNodes[0].parentNode === topNodes[topNodes.length - 1].parentNode);
                // Find siblings between #if and #else and between #else qnd #endif
                const ifSiblings = getSiblingsBetween(topNodes[0], topNodes[1]);
                const elseSiblings = elseMatchedNode instanceof MatchedNode ?
                    getSiblingsBetween(topNodes[1], topNodes[2]) :
                    [];
                // Evaluate #if condition
                let condition;
                try {
                    // Note: a condition which is truthy/falsy get converted to true/false
                    condition = (yield expressionEngine.evaluate(ifMatchedNode.expression, data));
                }
                catch (error) {
                    throw new YumError(1060, { error, data: { expression: ifMatchedNode.expression, data } });
                }
                // Update xmldom and ast
                const parent = topNodes[0].parentNode;
                for (const sibling of condition ? elseSiblings : ifSiblings) {
                    // Remove tags from ast where condition is not met
                    for (let i = this.children.length - 1; i >= 0; i--) {
                        const [matchedNode] = this.children[i].matchedNodes.values();
                        if (contains(sibling, matchedNode.node)) {
                            this.children.splice(i, 1);
                        }
                    }
                    // remove nodes from DOM
                    parent.removeChild(sibling);
                }
                // Remove #if, #else and #endif
                topNodes.forEach(node => { parent.removeChild(node); });
                // Process children where condition is met (others have been removed)
                for (let i = 0; i < this.children.length; i++) {
                    yield this.children[i].render(data);
                }
            }
        });
    }
}
IfTag.statement = '#if';
IfTag.blocks = ['#else', '#endif'];

/**
 * tagMap
 * Note: avoids circular references with YumTemplate
 */
// const tagMap = new Map<string, typeof AbstractTag>([
const tagMap = new Map([
    [ExpressionTag.statement, ExpressionTag],
    [EachTag.statement, EachTag],
    [IfTag.statement, IfTag]
]);

/**
 * TagParser
 */
class TagParser {
    /**
     * constructor
     * @param dom
     * @param options
     */
    constructor(dom, options = {}) {
        this._ast = [];
        this._stack = [];
        this._dom = dom;
        this._delimiters = (options.delimiters || constants.delimiters);
        this._current = this._ast;
    }
    /**
     * lexer
     * @private
     */
    get lexer() {
        // Note: we could have used a proper lexer like moo
        // but considering we rely on an expression parser like Jexl
        // our simple needs can be handled by a regular expression
        if (!(this._lexer instanceof RegExp)) {
            const statements = [];
            // Get registered *Tag blocks (TODO unicity?)
            for (const Tag of tagMap.values()) {
                statements
                    .push(escapeRegExp(Tag.statement), ...Tag.blocks.map(block => escapeRegExp(block)));
            }
            // Build the regular expression
            // TODO: we might have to escape certain characters like < and >
            this._lexer = new RegExp(escapeRegExp(this._delimiters.start) +
                '[ \t]*' +
                (statements.length ? `(?<statement>${statements.join('|')})?[ \t]*` : '') +
                `(?<expression>[^${escapeRegExp(this._delimiters.end.slice(0, 1))}]+)?` +
                '[ \t]*' +
                escapeRegExp(this._delimiters.end), 'g');
        }
        return this._lexer;
    }
    /**
     * _findTagInBlocks
     * @param statement
     * @private
     */
    _findTagInBlocks(statement) {
        for (const Tag of tagMap.values()) {
            if (Tag.blocks.includes(statement)) {
                return Tag;
            }
        }
    }
    /**
     * _parse
     * @param node
     * @private
     */
    _parse(node) {
        var _a;
        // if (node instanceof Text && this.lexer instanceof RegExp) {
        if (node.nodeType === 3 && this.lexer instanceof RegExp) {
            const matches = (_a = node.nodeValue) === null || _a === void 0 ? void 0 : _a.matchAll(this.lexer);
            if (matches) {
                for (const match of matches) {
                    const matchedNode = new MatchedNode(node, match);
                    const { statement, expression } = matchedNode;
                    const Tag = tagMap.get(statement || ExpressionTag.statement) ||
                        this._findTagInBlocks(statement);
                    if (Tag && (Tag.statement === ExpressionTag.statement) && expression) {
                        // If tag is a standalone expression
                        const tag = new Tag(matchedNode, this._current);
                        this._current.push(tag);
                    }
                    else if (Tag && (Tag.statement === statement)) {
                        // If tag corresponds to an opening statement, e.g. each or if, with or without expression
                        const tag = new Tag(matchedNode, this._current);
                        this._stack.push(tag);
                        this._current.push(tag);
                        if (Tag.blocks.length > 0) {
                            this._current = tag.children;
                        }
                    }
                    else if (Tag && Tag.blocks.includes(statement)) {
                        // If tag corresponds to a branch, e.g. else, or a closing statement in a block, e.g. endeach or endif
                        const isClosing = (Tag.blocks.indexOf(statement) === Tag.blocks.length - 1);
                        const main = isClosing ? this._stack.pop() : this._stack[this._stack.length - 1];
                        if (main instanceof Tag) {
                            main.matchedNodes.set(statement, matchedNode);
                            if (isClosing) {
                                this._current = main.parent;
                            }
                        }
                        else {
                            // Possibly main could be undefined
                            throw new YumError(2000); // TODO Review code
                        }
                    }
                    else {
                        throw new YumError(2000); // TODO Review code
                    }
                }
            }
        }
        if (node.hasChildNodes()) {
            const { childNodes } = node;
            // for (const child of childNodes) { - childNodes is not iterable
            for (let i = 0, { length } = childNodes; i < length; i++) {
                this._parse(childNodes.item(i));
            }
        }
    }
    /**
     * parse
     */
    parse() {
        this._parse(this._dom);
        // _stack should be empty and _current should be _ast;
        assert(this._stack.length === 0);
        assert(this._current === this._ast);
        return this._ast;
    }
}
// Assigned at runtime in TagParser.ts to resolve a circular dependency
// src/parts/TagParser.ts -> src/tags/tagMap.ts -> src/tags/EachTag.ts -> src/parts/TagParser.ts
AbstractTag.TagParser = TagParser;

/**
 * sanitizePptMarkup
 *
 *
 * @param xml
 * @param delimiters
 */
function sanitizePptMarkup(xml, delimiters) {
    const { start, end } = delimiters;
    const garbage = '<\\/a:t><\\/a:r><a:r\\s?[^>]*><a:t\\s?[^>]*>';
    let str = '';
    for (let i = 0; i < start.length - 1; i++) {
        str += `${escapeRegExp(start.charAt(i))}(${garbage})?`;
    }
    str += `${start.slice(-1)}[^${escapeRegExp(end.charAt(0))}]+${end.slice(0, 1)}`;
    for (let i = 1; i < start.length; i++) {
        str += `(${garbage})?${escapeRegExp(end.charAt(i))}`;
    }
    // const rx_find = /{(<\/a:t><\/a:r><a:r\s?[^>]*><a:t\s?[^>]*>)?{[^{}]+}(<\/a:t><\/a:r><a:r\s?[^>]*><a:t\s?[^>]*>)?}/;
    const rx_find = new RegExp(str);
    const rx_replace = /<\/?a:[^>]+>/g;
    let pos1 = 0, pos2 = 0, ret = '';
    while (xml.length > 0) {
        // Search for an expression in xml
        pos1 = xml.search(rx_find);
        if (pos1 > -1) {
            // find the end of the expression
            pos2 = xml.indexOf('}', pos1 + 1);
            pos2 = xml.indexOf('}', pos2 + 1);
            ret += xml.substring(0, pos1);
            ret += xml.substring(pos1, pos2 + '}'.length).replace(rx_replace, '');
            xml = xml.substring(pos2 + '}'.length);
        }
        else {
            ret += xml;
            xml = '';
        }
    }
    return ret;
}

/**
 * TemplatedPart
 */
class TemplatedPart extends AbstractPart {
    /**
     * constructor
     * @param name
     * @param type
     * @param xml
     * @param parent
     * @param options
     */
    constructor(name, type, xml, parent, options) {
        super(name, type, xml, parent, options);
        this.priority = 1;
        this._ast = [];
        this._ast = new TagParser(this._dom, options).parse();
    }
    /**
     * _preProcess
     * Note: executed by parent AbstractPart
     * @param xml
     */
    _preProcess(xml) {
        const { start, end } = this._options.delimiters;
        // TODO escape characters like < and >
        const pos1 = xml.indexOf(start.slice(0, 1));
        const pos2 = pos1 > -1 ? xml.indexOf(end.slice(-1), pos1 + 1) : -1;
        // No need to sanitize if we cannot find the first and last characters of delimiters
        if (this._name.startsWith('word/') && pos1 > 0 && pos2 > 0) {
            return sanitizeWordMarkup(xml, { start, end });
        }
        else if (this._name.startsWith('ppt/') && pos1 > 0 && pos2 > 0) {
            return sanitizePptMarkup(xml, { start, end });
        }
        else {
            // Do nothing by default
            return super._preProcess(xml);
        }
    }
    /**
     * render
     * @param data
     */
    render(data) {
        return __awaiter(this, void 0, void 0, function* () {
            if (this._done)
                return;
            for (const tag of this._ast) {
                yield tag.render(data);
            }
            this._done = true;
            // return this._dom; // for testing purpose only
        });
    }
}

/**
 * partMap
 * Note: avoids circular references with YumTemplate
 */
// const partMap = new Map<string, typeof AbstractPart>([
const partMap = new Map([
    // Word
    ['word/document.xml', TemplatedPart],
    ['word/footer.xml', TemplatedPart],
    ['word/header.xml', TemplatedPart],
    // Powerpoint
    ['ppt/slides/slide.xml', TemplatedPart],
    // Excel
    ['xl/sharedStrings.xml', TemplatedPart]
]);

class FileProxy {
}
class BlobProxy {
}
// eslint-disable-next-line @typescript-eslint/no-unused-vars
function saveAsProxy(blob, path) { return; } // TODO or throw?
const File = isNodeJS ? FileProxy : g.File;
const Blob = isNodeJS ? BlobProxy : g.Blob;
const saveAs = isNodeJS ? saveAsProxy : FileSaver.saveAs;

const CONTENT_TYPES = '[Content_Types].xml';
/**
 * YumTemplate
 */
class YumTemplate {
    /**
     * constructor
     * @param options
     */
    constructor(options = {}) {
        this._options = this._sanitizeOptions(options);
        this._parts = new Map();
        this._zip = new JSZip();
    }
    static resetTime() {
        YumTemplate.time = Date.now();
    }
    ;
    static showTime(message) {
        if (YumTemplate.instrument) {
            const t = Date.now();
            console.log(`${message}: ${t - YumTemplate.time} ms`);
            YumTemplate.time = t;
        }
    }
    ;
    // ----------------------------------
    // Cultures
    // ----------------------------------
    static registerCulture(locale, culture) {
        // Note: a registered culture can be replaced
        cultureMap.set(locale, culture);
    }
    // ----------------------------------
    // Parts
    // ----------------------------------
    // static registerPart(name: string, Part: typeof AbstractPart) {
    static registerPart(name, Part) {
        // Note: a registered part can be replaced
        partMap.set(name, Part);
    }
    // ----------------------------------
    // Tags
    // ----------------------------------
    // static registerTag(name: string, Tag: typeof AbstractTag) {
    static registerTag(name, Tag) {
        // Note: a registered tag can be replaced
        tagMap.set(name, Tag);
    }
    // ----------------------------------
    // Expression Engine
    // ----------------------------------
    static setExpressionEval(evaluate) {
        expressionEngine.setEval(evaluate);
    }
    /**
     * _sanitizeOptions
     *  options.delimiters
     *  options.locale
     * @param options
     * @private
     */
    _sanitizeOptions(options) {
        // TODO Possibly check invalid options and raise errors
        return {
            delimiters: (options.delimiters || constants.delimiters),
            locale: (options.locale || constants.locale)
        };
    }
    /**
     * _loadNodePath
     * @param path
     * @private
     */
    _loadNodePath(path) {
        return __awaiter(this, void 0, void 0, function* () {
            try {
                const handle = yield fs.promises.readFile(path);
                this._zip = yield JSZip.loadAsync(handle);
            }
            catch (error) {
                if (Object.prototype.hasOwnProperty.call(error, 'syscall')) {
                    throw new YumError(1010, { data: { path }, error });
                }
                else if (error.message.startsWith('Can\'t find end of central directory')) {
                    throw new YumError(1011, { data: { path }, error });
                }
                else {
                    throw new YumError(1012, { data: { path }, error });
                }
            }
        });
    }
    /**
     * _loadBrowserBlob
     * @param file
     * @private
     */
    _loadBrowserBlob(blob) {
        return __awaiter(this, void 0, void 0, function* () {
            try {
                this._zip = yield JSZip.loadAsync(blob);
            }
            catch (error) {
                console.log(error.message);
            }
        });
    }
    /**
     * _loadBrowserFile
     * @param file
     * @private
     */
    _loadBrowserFile(file) {
        return __awaiter(this, void 0, void 0, function* () {
            try {
                this._zip = yield JSZip.loadAsync(file);
            }
            catch (error) {
                console.log(error.message);
            }
        });
    }
    /**
     * load
     * @param handle
     */
    load(handle) {
        return __awaiter(this, void 0, void 0, function* () {
            YumTemplate.resetTime();
            if (isNodeJS && typeof handle === 'string') {
                yield this._loadNodePath(handle);
            }
            else if (!isNodeJS && handle instanceof File) {
                yield this._loadBrowserFile(handle);
            }
            else if (!isNodeJS && handle instanceof Blob) {
                yield this._loadBrowserBlob(handle);
            }
            else {
                throw new YumError(2000); // TODO review code + message
            }
            YumTemplate.showTime('Zip loaded');
        });
    }
    /**
     * _getPartRefs
     * @private
     */
    _getPartRefs() {
        var _a;
        return __awaiter(this, void 0, void 0, function* () {
            const ret = [];
            let xml;
            try {
                YumTemplate.resetTime();
                xml = (yield ((_a = this._zip.file(CONTENT_TYPES)) === null || _a === void 0 ? void 0 : _a.async('string'))) || '';
                YumTemplate.showTime('Content types loaded as xml');
                const dom = new DOMParser().parseFromString(xml, constants.mimeType);
                const nodes = dom.childNodes[isNodeJS ? 2 : 0].childNodes;
                for (let i = 0; i < nodes.length; i++) {
                    const node = nodes[i];
                    // We ignore nodes with nodeName === 'Default' nodes'
                    if (node.nodeName === 'Override') {
                        const { attributes } = node;
                        let name, type;
                        for (let j = 0; j < attributes.length; j++) {
                            const { nodeName, nodeValue } = attributes[j];
                            switch (nodeName) {
                                case 'PartName':
                                    name = (nodeValue || '').slice(1); // Remove /
                                    break;
                                case 'ContentType':
                                    type = nodeValue;
                                    break;
                                default:
                            }
                        }
                        if (name && type) {
                            const Part = partMap.get(name.replace(/\d+(.xml)$/, '$1'));
                            // Note: without registered part in YumTemplate.parts, no rendering
                            if (Part) {
                                ret.push({ name, type, Part });
                            }
                        }
                    }
                }
            }
            catch (error) {
                throw new YumError(1020, { data: { xml }, error });
            }
            YumTemplate.showTime('Content types read');
            return ret;
        });
    }
    /**
     * _loadParts
     * @private
     */
    _loadParts(partRefs) {
        return __awaiter(this, void 0, void 0, function* () {
            YumTemplate.resetTime();
            // Note: Priority order is not important here
            const promises = partRefs.map((ref) => __awaiter(this, void 0, void 0, function* () {
                const xml = yield this._zip.files[ref.name].async('text');
                this._parts.set(ref.name, new ref.Part(ref.name, ref.type, xml, this._parts, this._options));
            }));
            yield Promise.all(promises);
            YumTemplate.showTime('Content parts loaded');
        });
    }
    /**
     * render
     * @param data
     */
    render(data = {}) {
        return __awaiter(this, void 0, void 0, function* () {
            let ret;
            // List partRefs from [Content_Types].xml
            const partRefs = yield this._getPartRefs();
            // Load xml files into Parts including xml dom
            yield this._loadParts(partRefs);
            YumTemplate.resetTime();
            // Priority in increasing order (1 is highest, 10 is lowest)
            const parts = Array.from(this._parts.values())
                .sort((a, b) => (a.priority - b.priority));
            // Render and serialize each part
            for (const part of parts) {
                yield part.render(data);
                const xml = part.serialize();
                // For testing only
                if (part.name === 'word/document.xml' ||
                    part.name === 'ppt/slides/slide1.xml' ||
                    part.name === 'xl/sharedStrings.xml') {
                    ret = xml;
                }
                // Update zip stream
                this._zip.file(part.name, xml);
            }
            YumTemplate.showTime('Data merged');
            return ret; // For testing
        });
    }
    /**
     * saveAs
     * @param path
     */
    saveAs(path) {
        return __awaiter(this, void 0, void 0, function* () {
            YumTemplate.resetTime();
            if (isNodeJS) {
                const buf = yield this._zip.generateAsync({
                    type: 'nodebuffer',
                    streamFiles: true,
                    // compression: 'DEFLATE'
                });
                yield fs.promises.writeFile(path, buf);
            }
            else {
                const blob = yield this._zip.generateAsync({ type: 'blob' });
                saveAs(blob, path);
            }
            YumTemplate.showTime('Zip saved');
        });
    }
}
// ----------------------------------
// Instrumentation (can be removed)
// ----------------------------------
YumTemplate.instrument = false;
YumTemplate.time = 0;

exports.YumError = YumError;
exports.YumTemplate = YumTemplate;
exports.expressionEngine = expressionEngine;
